package it.inaf.oats.ia2.tap.tsm.rest;

import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import it.inaf.oats.ia2.tap.tsm.datalayer.ConnectionManager;
import it.inaf.oats.ia2.tap.tsm.datalayer.UpdateManager;
import it.inaf.oats.ia2.tap.tsm.model.Credentials;
import it.inaf.oats.ia2.tap.tsm.model.DbEntity;
import it.inaf.oats.ia2.tap.tsm.model.IDbEntitiesContainer;
import it.inaf.oats.ia2.tap.tsm.model.Status;
import it.inaf.oats.ia2.tap.tsm.model.Table;
import it.inaf.oats.ia2.tap.tsm.model.TapSchema;
import javax.ws.rs.core.Context;
import javax.ws.rs.Consumes;
import javax.ws.rs.Path;
import javax.ws.rs.Produces;
import javax.ws.rs.core.MediaType;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpSession;
import javax.ws.rs.FormParam;
import javax.ws.rs.POST;

/**
 * REST Web Service
 *
 * @author Sonia Zorba <zorba at oats.inaf.it>
 */
@Path("tapschema")
public class TapSchemaResource {

    private static final String PAGE_STATE = "pageState";

    // containers/entity names
    private static final String TAP_SCHEMA = "tapSchema";
    private static final String SCHEMA = "schema";
    private static final String TABLE = "table";
    private static final String COLUMN = "column";

    // models names
    private static final String SCHEMAS = "schemas";
    private static final String SCHEMA_VALUES = "schemaValues";
    private static final String TABLES = "tables";
    private static final String TABLE_VALUES = "tableValues";
    private static final String COLUMNS = "columns";
    private static final String COLUMN_VALUES = "columnValues";
    private static final String ADDABLES_MODAL = "addablesModal";

    private static final String[] CONTAINERS = {TAP_SCHEMA, SCHEMA, TABLE};

    private static final Gson STANDARD_GSON = new GsonBuilder().create();
    private static final Gson NULLABLE_GSON = new GsonBuilder().serializeNulls().create();

    class EditableField {

        String value;
        boolean isChanged;

        EditableField(String key, DbEntity container) {
            this.value = container.getValue(key);
            this.isChanged = container.isChanged(key);
        }
    }

    class DBEntitiesContainerStatus {

        boolean selectedToRemove;
        boolean hasAddables;
        boolean hasChildren;

        DBEntitiesContainerStatus(IDbEntitiesContainer container) {
            if (container != null) {
                hasAddables = !container.getAddables().isEmpty();
                hasChildren = !container.getAllEntities().isEmpty();
                if (hasChildren) {
                    selectedToRemove = container.getSelected().getStatus() == Status.TO_REMOVE;
                }
            }
        }
    }

    class PageModel {

        final Map<String, Object> elementsToUpdate;
        final Map<String, DBEntitiesContainerStatus> status;

        PageModel(Map<String, Object> elementsToUpdate) {
            this.elementsToUpdate = elementsToUpdate;

            status = new HashMap<String, DBEntitiesContainerStatus>();
            for (String container : CONTAINERS) {
                status.put(container, new DBEntitiesContainerStatus(getEntitiesContainer(container)));
            }
        }
    }

    class EntityItem {

        private final String title;
        private final boolean toRemove;
        private final boolean active;

        public EntityItem(DbEntity dbEntity, boolean active) {
            this.title = dbEntity.getName();
            this.toRemove = dbEntity.getStatus() == Status.TO_REMOVE;
            this.active = active;
        }
    }

    @Context
    private HttpServletRequest request;

    private final Credentials credentials = new Credentials("localhost", 3306, "root", "root");

    private TapSchema getTapSchema() {
        return getPageState().getTapSchema();
    }

    private SchemaPageState getPageState() {
        SchemaPageState pageState = null;
        HttpSession session = request.getSession();
        if (session.getAttribute(PAGE_STATE) == null) {

            Connection connection = null;
            try {
                connection = credentials.getConnection();
                pageState = new SchemaPageState(new TapSchema(connection, "vlkb_voinfo_schema"));
                session.setAttribute(PAGE_STATE, pageState);
            } catch (SQLException e) {
                e.printStackTrace(System.err);
            } finally {
                try {
                    if (connection != null) {
                        connection.close();
                    }
                } catch (SQLException e) {
                    e.printStackTrace(System.err);
                }
            }
        } else {
            pageState = (SchemaPageState) session.getAttribute(PAGE_STATE);
        }
        return pageState;
    }

    /**
     * Creates a new instance of TapSchemaResource
     */
    public TapSchemaResource() {
    }

    @POST
    @Path("model")
    @Produces(MediaType.APPLICATION_JSON)
    public String getPageModelJson() {
        PageModel pi = new PageModel(getNewModels());
        return NULLABLE_GSON.toJson(pi);
    }

    private Map<String, Object> getNewModels() {
        TapSchema tapSchema = getTapSchema();

        Map<String, Object> newModels = new HashMap<String, Object>();

        newModels.put(SCHEMAS, getAllEntities(tapSchema));
        newModels.put(TABLES, getAllEntities(tapSchema.getSelectedSchema()));
        newModels.put(COLUMNS, getAllEntities(tapSchema.getSelectedTable()));
        newModels.put(SCHEMA_VALUES, getEditablesFields(tapSchema.getSelectedSchema()));
        newModels.put(TABLE_VALUES, getEditablesFields(tapSchema.getSelectedTable()));
        newModels.put(COLUMN_VALUES, getEditablesFields(tapSchema.getSelectedColumn()));

        return newModels;
    }

    private String getPageModelUpdatesJson(String... excluded) {
        return NULLABLE_GSON.toJson(getPageModelUpdates(excluded));
    }

    private PageModel getPageModelUpdates(String... excluded) {
        return new PageModel(getPageState().getModelsChanged(getNewModels(), Arrays.asList(excluded)));
    }

    private ArrayList<EntityItem> getAllEntities(IDbEntitiesContainer container) {
        if (container == null) {
            return null;
        }

        ArrayList<EntityItem> entities = new ArrayList<EntityItem>();
        String selectedEntity = container.getSelected() == null ? null : container.getSelected().getName();
        for (DbEntity entity : container.getAllEntities()) {
            entities.add(new EntityItem(entity, entity.getName().equals(selectedEntity)));
        }
        return entities;
    }

    private HashMap<String, EditableField> getEditablesFields(DbEntity container) {
        if (container == null) {
            return null;
        }

        HashMap<String, EditableField> fields = new HashMap<String, EditableField>();
        for (String key : container.getSupportedValues()) {
            fields.put(key, new EditableField(key, container));
        }
        return fields;
    }

    private IDbEntitiesContainer getEntitiesContainer(String name) {
        TapSchema tapSchema = getTapSchema();
        if (name.equals(TAP_SCHEMA)) {
            return tapSchema;
        }
        if (name.equals(SCHEMA)) {
            return tapSchema.getSelectedSchema();
        }
        if (name.equals(TABLE)) {
            return tapSchema.getSelectedTable();
        }
        return null;
    }

    private DbEntity getEditablesContainer(String name) {
        TapSchema tapSchema = getTapSchema();
        if (name.equals(SCHEMA)) {
            return tapSchema.getSelectedSchema();
        }
        if (name.equals(TABLE)) {
            return tapSchema.getSelectedTable();
        }
        if (name.equals(COLUMN)) {
            return tapSchema.getSelectedColumn();
        }
        return null;
    }

    @POST
    @Path("select")
    public String select(@FormParam("containerName") String containerName, @FormParam("entityName") String entityName) {
        IDbEntitiesContainer container = getEntitiesContainer(containerName);
        container.select(entityName);

        if (containerName.equals(TABLE)) {
            return getPageModelUpdatesJson(COLUMNS);
        }
        return getPageModelUpdatesJson();
    }

    class RemoveColumnResult {

        PageModel pageModelUpdates;
        boolean toRemove;
    }

    @POST
    @Path("remove")
    public String remove(@FormParam("containerName") String containerName, @FormParam("entityName") String entityName) {
        IDbEntitiesContainer container = getEntitiesContainer(containerName);
        container.removeEntity(entityName);

        if (containerName.equals(TABLE)) {
            RemoveColumnResult result = new RemoveColumnResult();
            result.pageModelUpdates = getPageModelUpdates(COLUMNS);
            result.toRemove = ((Table) container).getColumn(entityName).getStatus() == Status.TO_REMOVE;
            return NULLABLE_GSON.toJson(result);
        }

        return getPageModelUpdatesJson();
    }

    @POST
    @Path("addables")
    public String getAddables(@FormParam("containerName") String containerName) {
        IDbEntitiesContainer container = getEntitiesContainer(containerName);
        return STANDARD_GSON.toJson(container.getAddables());
    }

    class AddParams {

        String containerName;
        List<String> entities;
    }

    @POST
    @Path("update")
    public String update() throws SQLException {
        TapSchema tapSchema = getTapSchema();
        (new UpdateManager(credentials, tapSchema)).update();
        return "updated";
    }
    
    @POST
    @Path("add")
    @Consumes(MediaType.APPLICATION_JSON)
    @Produces(MediaType.APPLICATION_JSON)
    public String add(String params) throws SQLException {
        final AddParams addParams = STANDARD_GSON.fromJson(params, AddParams.class);

        (new ConnectionManager(credentials) {

            @Override
            public void use(Connection connection) throws SQLException {
                getEntitiesContainer(addParams.containerName).loadEntities(connection, addParams.entities);
            }
        }).connect();

        return getPageModelUpdatesJson();
    }

    class EditParams {

        String container;
        String key;
        String value;
    }

    @POST
    @Path("edit")
    @Consumes(MediaType.APPLICATION_JSON)
    @Produces(MediaType.APPLICATION_JSON)
    public String edit(String params) throws SQLException {
        EditParams editParams = STANDARD_GSON.fromJson(params, EditParams.class);
        return getEditablesContainer(editParams.container).updateValue(editParams.key, editParams.value) + "";
    }

    @POST
    @Path("undoremove")
    @Produces(MediaType.APPLICATION_JSON)
    public String undoremove(@FormParam("containerName") String containerName) throws SQLException {
        getEntitiesContainer(containerName).getSelected().setStatus(Status.SELECTED_SAVED);
        return getPageModelUpdatesJson();
    }

    @POST
    @Path("searchUCD")
    @Produces(MediaType.APPLICATION_JSON)
    public String searchUCD(@FormParam("description") String description) throws Exception {
        SearchUCDDialog dialog = getPageState().getSearchUCDDialog();
        if (description == null || description.equals("")) {
            dialog.setDefault();
        } else {
            dialog.search(description);
        }

        return NULLABLE_GSON.toJson(dialog);
    }
    
    @POST
    @Path("selectUCD")
    public String selectUCD(@FormParam("index") String indexStr) {
        SearchUCDDialog dialog = getPageState().getSearchUCDDialog();
        
        if(indexStr == null || indexStr.equals("")) {
            dialog.selectUCD();
        } else {
            dialog.selectUCD(Integer.parseInt(indexStr));
        }

        return dialog.getSelectedUCD();
    }
    
    @POST
    @Path("explainUCD")
    public String explainUCD(@FormParam("index") String indexStr) throws Exception {
        SearchUCDDialog dialog = getPageState().getSearchUCDDialog();
        
        if(indexStr == null || indexStr.equals("")) {
            return dialog.explain();
        }
        
        return dialog.explain(Integer.parseInt(indexStr));
    }
}
